#![allow(non_upper_case_globals)]
#![no_std]

extern crate alloc;
use alloc::boxed::Box;
use alloc::vec;
use alloc::vec::Vec;
use core::ops;
use lazy_static::lazy_static;

macro_rules! BM {
    ($exp:expr) => {
        (1 << $exp) - 1;
    };
}

// The theory throughout this codebase is to use raw arrays over Vec whenever possible if it only
// costs a couple of bytes.

// fixed length extra bits
const fleb: [u8; 32] = [
    0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0,
    /* unused */ 0, 0, /* impossible */ 0,
];
const flebmsk: [u8; 32] = [
    0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 3, 3, 3, 3, 7, 7, 7, 7, 15, 15, 15, 15, 31, 31, 31, 31, 0,
    0, 0, 0,
];
// fixed distance extra bits
const fdeb: [u8; 32] = [
    0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13,
    13, /* unused */ 0, 0,
];
const fdebmsk: [u16; 32] = [
    0, 0, 0, 0, 1, 1, 3, 3, 7, 7, 15, 15, 31, 31, 63, 63, 127, 127, 255, 255, 511, 511, 1023, 1023,
    2047, 2047, 4095, 4095, 8191, 8191, 0, 0,
];
// code length index map
const clim: [u8; 19] = [
    16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15,
];

// from extrabits (get maps)
fn freb(eb: &[u8], r: &mut [u32], mut st: u16) -> [u16; 31] {
    let mut b = [0u16; 31];
    for i in 0..31 {
        st += 1 << eb[i - 1];
        b[i] = st;
    }
    // numbers here are at max 18 bits
    for i in 1..30 {
        for j in b[i]..b[i + 1] {
            r[j as usize] = (((j - b[i]) << 5) | i as u16) as u32;
        }
    }
    b
}

// h(uff)map base
fn hmb(cd: &[u8]) -> [u16; 16] {
    // u8 "map": index -> # of codes with bit length = index
    let mut l = [0u8; 16];
    // length of cd must be 288 (total # of codes)
    for i in 0..cd.len() {
        l[cd[i] as usize] += 1;
    }
    // u16 "map": index -> minimum code for bit length = index
    let mut le = [0u16; 16];
    for i in 0..15 {
        le[i] = (le[i - 1] + l[i] as u16) << 1;
    }
    le
}

// h(uff)map
fn hmap(cd: &[u8]) -> [u16; 288] {
    let mut le = hmb(cd);
    let mut co = [0u16; 288];
    for i in 0..cd.len() {
        co[i] = rev[le[cd[i] as usize] as usize] >> (15 - cd[i]);
        le[cd[i] as usize] += 1;
    }
    co
}

// h(uff)r(everse)map
fn hrmap(cd: &[u8], mb: u8) -> Box<[u16]> {
    let mut le = hmb(cd);
    let mut co = vec![0u16; 1 << mb].into_boxed_slice();
    let rvb = 15 - mb;
    for i in 0..cd.len() {
        // ignore 0 lengths
        if cd[i] != 0 {
            // num encoding both symbol and bits read
            let sv = (i << 4) as u16 | cd[i] as u16;
            // free bits
            let r = mb - cd[i];
            // start value
            let v = le[cd[i] as usize] << r;
            le[cd[i] as usize] += 1;
            // m is end value
            let m = v + (1 << r);
            for i in v..m {
                // every 16 bit value starting with the code yields the same result
                co[rev[i as usize] as usize >> rvb] = sv;
            }
        }
    }
    co
}

#[inline]
fn bits(d: &[u8], p: usize, m: u16) -> u16 {
    let o = p >> 3;
    ((d[o] as u16 | ((d[(o + 1)] as u16) << 8)) >> (p & 7)) & m
}

#[inline]
fn bits16(d: &[u8], p: usize, m: u32) -> u32 {
    let o = p >> 3;
    ((d[o] as u32 | ((d[(o + 1)] as u32) << 8) | ((d[(o + 2)] as u32) << 16)) >> (p & 7)) & m
}

#[inline]
fn shft(p: usize) -> usize {
    (p >> 3) + (p & 7 != 0) as usize
}

lazy_static! {
    // fl, revfl
    static ref fl: ([u16; 31], [u32; 261]) = {
        let mut r = [0u32; 261];
        let mut f = freb(&fleb, &mut r, 2);
        f[28] = 258;
        r[258] = 28;
        (f, r)
    };
    // fd, revfd
    static ref fd: ([u16; 31], [u32; 32769]) = {
        let mut r = [0u32; 32769];
        (freb(&fdeb, &mut r, 0), r)
    };
    // 15 bit reverse index map
    static ref rev: [u16; 32768] = {
        let mut r = [0u16; 32768];
        for i in 0..32768 {
            let mut x: u32 = ((i & 0xAAAAAAAAu32) >> 1) | ((i & 0x55555555u32) << 1);
            x = ((x & 0xCCCCCCCCu32) >> 2) | ((x & 0x33333333u32) << 2);
            x = ((x & 0xF0F0F0F0u32) >> 4) | ((x & 0x0F0F0F0Fu32) << 4);
            r[i as usize] = ((((x & 0xFF00FF00u32) >> 8) | ((x & 0x00FF00FFu32) << 8)) >> 1) as u16;
        }
        r
    };
    static ref flt: [u8; 288] = {
        let mut o = [8u8; 288];
        for i in 144..256 { o[i] = 9; }
        for i in 256..280 { o[i] = 7; }
        o
    };
    static ref fdt: [u8; 32] = [5; 32];
}

pub trait Writable: ops::DerefMut<Target = [u8]> {
    fn need_bytes(&mut self, bt: usize) -> bool;
}

impl Writable for Vec<u8> {
    fn need_bytes(&mut self, bt: usize) -> bool {
        self.reserve(bt);
        true
    }
}

impl Writable for &mut [u8] {
    // todo: maintain size
    fn need_bytes(&mut self, _bt: usize) -> bool {
        true
    }
}

struct InflateState<'a> {
  lmap: &'a [u16],
  dmap: &'a [u16],
  lbits: u8,
  dbits: u8,
  last: bool,
  pos: usize,
  bt: usize,
  lsch: bool,
  bule: bool
}

// fn inflt(dat: &[u8], buf: &dyn Writable<Output = u8>, st: &mut InflateState) {
    
// }